import jax
from jax.flatten_util import ravel_pytree
from .core import (
    sample_dist,
    log_prob_dist,
    fill_in_bijector,
    inverse_transform_dist,
    transform_tree,
)

import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions


class MCMC:
    def __init__(self, prior, bijector={}, log_likelihood_fn=None):
        self.prior = prior
        self.bijector = fill_in_bijector(bijector, self.prior)
        self.log_likelihood_fn = log_likelihood_fn

        self.approx_normal_prior = inverse_transform_dist(self.prior, self.bijector)

        dummy_params = self.init(jax.random.PRNGKey(0))
        _, self.unravel_fn = ravel_pytree(dummy_params)

    def init(self, seed):
        return sample_dist(self.approx_normal_prior, seed)

    def log_joint(self, params, outputs, inputs):
        p_log_prob = log_prob_dist(self.approx_normal_prior, params)
        constrained_params = transform_tree(params, self.bijector)
        log_likelihood = self.log_likelihood_fn(latent_sample=constrained_params, outputs=outputs, inputs=inputs)
        return p_log_prob + log_likelihood

    def get_nuts_kernel(self, outputs, inputs, step_size=1e-3, **kwargs):
        def log_joint_fn(params):
            log_joint = jax.tree_util.Partial(self.log_joint, outputs=outputs, inputs=inputs)
            return log_joint(self.unravel_fn(params))

        return tfp.mcmc.NoUTurnSampler(log_joint_fn, step_size=step_size, **kwargs)

    def sample(self, seed, init_params_pytree, kernel, n_samples=100, n_burnin=100, trace_fn=None):
        if trace_fn is None:
            trace_fn = lambda _, results: results.target_log_prob
        init_params, _ = ravel_pytree(init_params_pytree)

        states, *other_results = tfp.mcmc.sample_chain(
            num_results=n_samples,
            current_state=init_params,
            trace_fn=trace_fn,
            kernel=kernel,
            num_burnin_steps=n_burnin,
            seed=seed,
        )

        states_pytree = jax.vmap(self.unravel_fn)(states)
        constrained_states = transform_tree(states_pytree, self.bijector)
        return constrained_states, other_results



import jax
import jax.numpy as jnp
import distrax

mean = jnp.array([0.1, 0.2])
scale = jnp.array([0.4, 0.5])
key = jax.random.PRNGKey(1)

dist_uni = distrax.Normal(mean, scale)
trans_uni = distrax.Transformed(dist_uni, distrax.Sigmoid())
trans_uni_sample, log_prob = trans_uni.sample_and_log_prob(seed=key)
print(f"{trans_uni_sample=}, {log_prob=}, {log_prob.sum()=}")

# dist_bi = distrax.MultivariateNormalDiag(mean, scale)
# trans_bi = distrax.Transformed(dist_bi, distrax.Sigmoid(event_ndims_in=2))
# trans_bi_sample, log_prob = trans_bi.sample_and_log_prob(seed=key)
# print(f"{trans_bi_sample=}, {log_prob}")

dist_bi = distrax.MultivariateNormalDiag(mean, scale)
transform = jax.nn.sigmoid
trans_bi_sample, log_prob = dist_bi.sample_and_log_prob(seed=key)
log_jac = jnp.log(jnp.linalg.det(jax.jacfwd(transform)(trans_bi_sample)))
print(f"{transform(trans_bi_sample)=}, {log_prob - log_jac}")



######### Avoid TypeCheck Warning #########
import logging

logger = logging.getLogger()


class CheckTypesFilter(logging.Filter):
    def filter(self, record):
        return "check_types" not in record.getMessage()


logger.addFilter(CheckTypesFilter())
############################################

import jax
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions
tfb = tfp.bijectors

from .utils import seeds_like


class Posterior:
    def __init__(self, posterior, approx_normal_prior, bijector):
        assert isinstance(posterior, tfd.Distribution)
        self.posterior = posterior
        self.approx_normal_prior = approx_normal_prior
        self.bijector = bijector

    def sample(self, seed, sample_shape=()):
        sample = self.posterior.sample(seed=seed, sample_shape=sample_shape)
        unravel_fn = get_normal_posterior_size_and_unravel_fn(self.approx_normal_prior)[1]

        def f(x):
            return transform_tree(unravel_fn(x), self.bijector)

        for _ in range(len(sample_shape)):
            f = jax.vmap(f)
        return f(sample)

    def log_prob(self, sample, sample_shape=()):
        def log_prob_per_sample(sample_tree):
            approx_normal_sample_tree = inverse_transform_tree(sample_tree, self.bijector)
            approx_normal_sample, _ = ravel_pytree(approx_normal_sample_tree)
            normal_log_prob = self.posterior.log_prob(approx_normal_sample)
            jacobian_tree = jax.tree_map(
                lambda sample, bijector: bijector.inverse_log_det_jacobian(sample), sample_tree, self.bijector
            )
            jacobian = ravel_pytree(jacobian_tree)[0].sum()
            return normal_log_prob + jacobian

        f = log_prob_per_sample
        for _ in range(len(sample_shape)):
            f = jax.vmap(f)
        return f(sample)

    def prob(self, sample, sample_shape):
        log_prob = self.log_prob(sample, sample_shape)
        return jax.tree_map(jnp.exp, log_prob)


def fill_in_bijector(bijector, prior):
    for key in prior:
        if key not in bijector:
            bijector[key] = tfb.Identity()
    return bijector


def inverse_transform_dist(dist_pytree, bijector_pytree):
    is_leaf = lambda x: isinstance(x, tfd.Distribution)
    return jax.tree_map(
        lambda dist, bijector: tfd.TransformedDistribution(dist, tfb.Invert(bijector)),
        dist_pytree,
        bijector_pytree,
        is_leaf=is_leaf,
    )


def sample_dist(dist_pytree, seed, sample_shape=()):
    is_leaf = lambda dist: isinstance(dist, tfd.Distribution)
    seeds = seeds_like(dist_pytree, seed, is_leaf=is_leaf)
    samples = jax.tree_map(
        lambda dist, seed: dist.sample(seed=seed, sample_shape=sample_shape), dist_pytree, seeds, is_leaf=is_leaf
    )
    return samples


def log_prob_dist(dist_pytree, sample_pytree):
    def is_leaf_dist(dist):
        return isinstance(dist, tfd.Distribution)

    def log_prob(dist, sample):
        if isinstance(dist, tfd.Distribution):
            return dist.log_prob(sample)
        elif callable(dist):
            var_names = dist.__code__.co_varnames
            args = {var_name: sample_pytree[var_name] for var_name in var_names}
            return dist(**args).log_prob(sample)
        else:
            raise ValueError(f"Unknown type {type(dist)}")

    log_probs = jax.tree_map(
        lambda dist, sample: log_prob(dist, sample), dist_pytree, sample_pytree, is_leaf=is_leaf_dist
    )

    return ravel_pytree(log_probs)[0].sum()  # sum over batch dimension


def transform_tree(pytree, bijector_pytree):
    return jax.tree_map(lambda param, bijector: bijector(param), pytree, bijector_pytree)


def inverse_transform_tree(pytree, bijector_pytree):
    print(bijector_pytree)
    return jax.tree_map(lambda param, bijector: bijector.inverse(param), pytree, bijector_pytree)


def transform_dist_params(dist, ordered_posterior_bijectors):
    params, treedef = jax.tree_flatten(dist)
    params = jax.tree_map(lambda param, bijector: bijector(param), params, ordered_posterior_bijectors)
    return jax.tree_unflatten(treedef, params)


def get_normal_posterior_size_and_unravel_fn(approx_normal_prior):
    seed = jax.random.PRNGKey(0)
    samples = sample_dist(approx_normal_prior, seed=seed, sample_shape=())
    # for name, sample in samples.items():
    #     assert (
    #         sample.ndim < 2
    #     ), f"'{name}' returned {sample.ndim} dimensional sample after transformation but should be 1d or scalar"
    array, unravel_fn = ravel_pytree(samples)
    return len(array), unravel_fn


def get_mean_field(approx_normal_prior, ordered_posterior_bijectors=None):
    if ordered_posterior_bijectors is None:
        ordered_posterior_bijectors = [tfb.Identity(), tfb.Exp()]
    size, unravel_fn = get_normal_posterior_size_and_unravel_fn(approx_normal_prior)
    return (
        tfd.MultivariateNormalDiag(loc=jnp.zeros(size), scale_diag=jnp.ones(size)),
        unravel_fn,
        ordered_posterior_bijectors,
    )


def get_low_rank(approx_normal_prior, rank=1, ordered_posterior_bijectors=None):
    if ordered_posterior_bijectors is None:
        ordered_posterior_bijectors = [tfb.Identity(), tfb.Exp(), tfb.Identity()]
    size, unravel_fn = get_normal_posterior_size_and_unravel_fn(approx_normal_prior)
    scale_perturb_factor = jnp.ones((size, rank))
    return (
        tfd.MultivariateNormalDiagPlusLowRank(
            loc=jnp.zeros(size), scale_diag=jnp.ones(size), scale_perturb_factor=scale_perturb_factor
        ),
        unravel_fn,
        ordered_posterior_bijectors,
    )


def get_full_rank(approx_normal_prior, ordered_posterior_bijectors=None):
    if ordered_posterior_bijectors is None:
        ordered_posterior_bijectors = [tfb.Identity(), tfb.Identity()]
    size, unravel_fn = get_normal_posterior_size_and_unravel_fn(approx_normal_prior)
    return (
        tfd.MultivariateNormalTriL(loc=jnp.zeros(size), scale_tril=jnp.eye(size)),
        unravel_fn,
        ordered_posterior_bijectors,
    )


def check_distribution_zero_batch(dist_pytree):
    is_leaf = lambda x: isinstance(x, tfd.Distribution)

    def len_of_batch(dist):
        if isinstance(dist, tfd.Distribution):
            return len(dist.batch_shape)
        elif callable(dist):
            return 0

    batch_lens = jax.tree_map(lambda dist: len_of_batch(dist), dist_pytree, is_leaf=is_leaf)
    assert sum(jax.tree_leaves(batch_lens)) == 0, "The prior distributions must not have any batch dimension."



import logging

logger = logging.getLogger()


class CheckTypesFilter(logging.Filter):
    def filter(self, record):
        return "check_types" not in record.getMessage()


logger.addFilter(CheckTypesFilter())

import numpy as np
import jax
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions
tfb = tfp.bijectors

from .core import seeds_like, inverse_transform_dist, Posterior, transform_dist_params


class ADLaplace:
    def __init__(self, prior, bijector, get_likelihood=None):
        self.prior = prior
        self.bijectors = bijector
        self.get_likelihood = get_likelihood

        self.approx_normal_distributions = inverse_transform_dist(prior, bijector)

    def init(self, seed):
        seeds = seeds_like(self.prior, seed)
        normal_prior = inverse_transform_dist(self.prior, self.bijector)
        params = jax.tree_map(
            lambda seed, dist: dist.sample(seed=seed),
            seeds,
            normal_prior,
        )
        for param in params.values():
            assert param.ndim <= 1, "params must be scalar or vector"
        return params

    def loss_fn(self, params, outputs, inputs=None):
        transformed_params = jax.tree_map(lambda param, bijector: bijector.forward(param), params, self.bijectors)
        prior_log_probs = jax.tree_map(
            lambda param, dist: dist.log_prob(param),
            params,
            self.approx_normal_distributions,
        )

        def likelihood_log_prob(params, outputs, inputs):
            likelihood = self.get_likelihood(params, inputs)
            return likelihood.log_prob(outputs)

        likelihood_log_probs = jax.vmap(likelihood_log_prob, in_axes=(None, 0, 0))(transformed_params, outputs, inputs)
        loss = -(likelihood_log_probs.sum() + sum(jax.tree_leaves(prior_log_probs)))
        return loss

    def apply(self, params, outputs, inputs=None):
        precision = jax.hessian(self.loss_fn)(params, outputs, inputs)
        event_shapes = jax.tree_map(lambda _, dist: dist.event_shape, self.guide, self.prior)
        return Posterior(params, precision, self.bijectors, event_shapes)


# class Posterior:
#     def __init__(self, params, precision, bijectors, event_shapes):
#         self.params = jax.tree_map(lambda x: x, params)
#         self.guide = jax.tree_structure(self.params)
#         self.precision = jax.tree_map(lambda x: x, precision)
#         self.bijectors = jax.tree_map(lambda x: x, bijectors)
#         self.event_shapes = jax.tree_map(lambda x: x, event_shapes)

#         self.size = jax.tree_map(lambda param: param.size, self.params)
#         self.cumsum_size = np.cumsum(jax.tree_leaves(self.size))[:-1]

#     def untree_precision(self):
#         precision_matrix = []
#         for start in self.precision:
#             precision_matrix.append([])
#             for end in self.precision[start]:
#                 element = self.precision[start][end].reshape((self.size[start], self.size[end]))
#                 precision_matrix[-1].append(element)
#         return jnp.block(precision_matrix)

#     def get_normal_dist(self, keys):
#         precision_matrix = self.untree_precision()
#         covariance_matrix = jnp.linalg.inv(precision_matrix)
#         params = jax.tree_map(lambda x: x, self.params)
#         rm_keys = set(params) - set(keys)
#         for key in rm_keys:
#             params.pop(key)
#         loc, untree_fn = ravel_pytree(params)
#         flat_size, unravel = ravel_pytree(self.size)
#         cumsum_size = unravel(np.cumsum([0] + flat_size.tolist())[:-1].astype(np.int32))
#         idx_list = []
#         for key in keys:
#             idx_list.extend(range(cumsum_size[key], cumsum_size[key] + self.params[key].size))
#         idx_list = np.array(idx_list)
#         dist = tfd.MultivariateNormalFullCovariance(loc, covariance_matrix[jnp.ix_(idx_list, idx_list)])
#         return dist, untree_fn

#     def sample(self, seed, sample_shape=()):
#         dist, untree_fn = self.get_normal_dist(self.params.keys())
#         normal_sample = dist.sample(sample_shape=sample_shape, seed=seed)
#         f = untree_fn
#         for _ in range(len(sample_shape)):  # vectorized transform
#             f = jax.vmap(f)
#         split_sample = f(normal_sample)
#         return jax.tree_map(lambda sample, bijector: bijector.forward(sample), split_sample, self.bijectors)

#     # To Do: This function is problamatic so, correct it with tricks like vectorized transform in "sample" method
#     # inverse_log_det_jacobian event_shape is merely doing summation, need to check if it is correct.
#     def log_prob(self, sample, sample_shape=()):
#         def log_prob_per_sample(sample):
#             bijectors = jax.tree_map(lambda x: x, self.bijectors)
#             rm_keys = set(bijectors) - set(sample)
#             for key in rm_keys:
#                 bijectors.pop(key)
#             normal_sample = jax.tree_map(lambda sample, bijector: bijector.inverse(sample), sample, bijectors)
#             log_jacobian = jax.tree_map(
#                 lambda samp, bijector: bijector.inverse_log_det_jacobian(samp), sample, bijectors
#             )
#             log_jacobian = ravel_pytree(log_jacobian)[0].sum()
#             flat_sample, _ = ravel_pytree(normal_sample)
#             dist, _ = self.get_normal_dist(sample.keys())
#             normal_log_prob = dist.log_prob(flat_sample)
#             return normal_log_prob + log_jacobian

#         f = log_prob_per_sample
#         for _ in range(len(sample_shape)):
#             f = jax.vmap(f)
#         return f(sample)

#     def sample_and_log_prob(self, seed, sample_shape=()):
#         samples = self.sample(seed, sample_shape)
#         return samples, self.log_prob(samples, sample_shape)



import jax

from .core import (
    inverse_transform_dist,
    transform_tree,
    transform_dist_params,
    log_prob_dist,
    get_full_rank,
    get_low_rank,
    get_mean_field,
    fill_in_bijector,
)

from .core import Posterior
from .utils import initialize_params

import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions


class ADVI:
    def __init__(
        self,
        prior,
        bijector={},
        log_likelihood_fn=None,
        vi_type="mean_field",
        rank=None,
        ordered_posterior_bijectors=None,
    ):
        """Automatic Differentiation Variational Inference
        A model class that implements the ADVI algorithm.

        Args:
            prior (dict): A dictionary of prior distributions.
            bijector (dict): A dictionary of bijectors. The keys should be the same as the keys in `prior`.
            log_likelihood_fn (function): A function that returns log likelihood of outputs. The function should take two arguments: `likelihood_params` and `inputs`.
                                          example:
                                            prior = {"p_of_head": tfd.Beta(0.5, 0.5)}
                                            seed = jax.random.PRNGKey(0)
                                            likelihood_params = sample_dist(prior, seed) # likelihood_params = {"p_of_head": 0.5}
                                            def coin_toss_log_likelihood(likelihood_params, outputs, inputs):
                                                p_of_head = likelihood_params["p_of_head"]
                                                return tfd.Bernoulli(probs=p_of_head).log_prob(outputs)

            vi_type (str, optional): type of variational inference ("mean_field", "full_rank", "low_rank"). Defaults to "mean_field".
            rank (int, optional): Rank of posterior covariance matrix in case where `vi_type` is "low_rank". Defaults to None.
            ordered_posterior_bijectors (dict, optional): A dictionary of bijectors for posterior parameters. `ordered` here means ordered as per the alphabetical order. Defaults to None.
                                                  example:
                                                    In case of "mean_field" variational inference, the posterior is a `tfd.MultivariateNormalDiag`.
                                                    So, the ordered_posterior_bijectors can be [tfb.Identity(), tfb.Exp()] corresponding to `loc` and `scale_diag`.
        """

        self.prior = prior
        self.bijector = fill_in_bijector(bijector, self.prior)

        assert self.bijector.keys() == self.prior.keys(), "The keys in `prior` and `bijector` must be the same."

        assert (vi_type == "low_rank") == (
            rank is not None
        ), "`rank` must be specified only if `vi_type` is `low_rank`."

        self.log_likelihood_fn = log_likelihood_fn
        self.vi_type = vi_type

        self.approx_normal_prior = inverse_transform_dist(self.prior, self.bijector)

        if vi_type == "mean_field":
            self.posterior, self.unravel_fn, self.posterior_params_bijector = get_mean_field(
                self.approx_normal_prior, ordered_posterior_bijectors
            )
        elif vi_type == "low_rank":
            self.posterior, self.unravel_fn, self.posterior_params_bijector = get_low_rank(
                self.approx_normal_prior, rank, ordered_posterior_bijectors
            )
        elif vi_type == "full_rank":
            self.posterior, self.unravel_fn, self.posterior_params_bijector = get_full_rank(
                self.approx_normal_prior, ordered_posterior_bijectors
            )

    def init(self, seed, initializer=jax.nn.initializers.normal(stddev=1.0)):
        return {"posterior": initialize_params(self.posterior, seed, initializer)}

    def loss_fn(self, params, outputs, inputs, full_data_size, seed, n_samples=1):
        posterior = params["posterior"]
        posterior = transform_dist_params(posterior, self.posterior_params_bijector)

        def loss_fn_per_sample(seed):
            sample = posterior.sample(seed=seed)
            q_log_prob = posterior.log_prob(sample)
            sample_tree = self.unravel_fn(sample)
            p_log_prob = log_prob_dist(self.approx_normal_prior, sample_tree)
            transformed_sample_tree = transform_tree(sample_tree, self.bijector)
            log_likelihood = self.log_likelihood_fn(
                latent_sample=transformed_sample_tree, outputs=outputs, inputs=inputs, **params
            )
            log_likelihood = (log_likelihood / len(outputs)) * full_data_size  # normalize by outputs size
            return q_log_prob - p_log_prob - log_likelihood

        seeds = jax.random.split(seed, n_samples)
        return jax.vmap(loss_fn_per_sample)(seeds).mean()

    def apply(self, params):
        posterior = params["posterior"]
        posterior = transform_dist_params(posterior, self.posterior_params_bijector)
        return Posterior(posterior, self.approx_normal_prior, self.bijector)



import jax
from jax.flatten_util import ravel_pytree
import optax


def train_fn(loss_fn, params, optimizer, n_epochs, seed, return_args={}):
    value_and_grad_fn = jax.value_and_grad(loss_fn)
    state = optimizer.init(params)

    @jax.jit
    def one_step(params_and_state, seed):
        params, state = params_and_state
        loss, grads = value_and_grad_fn(params, seed=seed)
        updates, state = optimizer.update(grads, state)
        params = optax.apply_updates(params, updates)
        return (params, state), (loss, params)

    seeds = jax.random.split(seed, n_epochs)
    (params, states), (losses, params_list) = jax.lax.scan(one_step, (params, state), xs=seeds)
    return_dict = {"params": params}
    for key in return_args:
        return_dict[key] = locals()[key]
    return return_dict


def initialize_params(params, seed, initializer):
    """Initialize parameters with a given initializer.

    Args:
        seed: JAX PRNGKey
        params: parameters to initialize
        initializer: one of the jax.nn.initializers or a callable that takes a PRNGKey and shape and returns DeviceArray.
                     example: jax.nn.initializers.normal(stddev=0.1)
                     example: lambda seed, shape: jax.random.uniform(seed, shape, minval=0, maxval=1)
                     example: lambda seed, shape: jnp.zeros(shape)

    Returns:
        DeviceArray: initialized parameters
    """
    values, unravel_fn = ravel_pytree(params)
    random_values = initializer(seed, (len(values),))
    return unravel_fn(random_values)


def seeds_like(params, seed, is_leaf=None):
    """Generate seeds for a tree of parameters.

    Args:
        params: parameters.
        seed: JAX PRNGKey.
        is_leaf: same as `jax.tree_map`.

    Returns:
        seeds: JAX PRNGKeys for the tree of parameters.
    """
    values, treedef = jax.tree_flatten(params, is_leaf=is_leaf)
    return jax.tree_unflatten(treedef, jax.random.split(seed, len(values)))
